/*
 * MemAccessMonitor.cpp
 *
 *  Created on: 14/06/2009
 *      Author: Gil Kulish
 */

#include "MemAccessMonitor.h"
#include "ProcessorsManager.h"
#include <list>
#include <fstream>
#include <iostream>
using namespace std;

pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
/**
 Caches work in "chucks", each time a "line" is fetched into the memory (and not a single address).
 so this function transforms an address in the memory to the address of the memory line.
 */
PTR_TYPE MemAccessMonitor::getMemLineAddressByMemAddress(PTR_TYPE mem) {
	return (mem & MEM_MASK);
}

MemAccessMonitor::MemAccessMonitor() {
	// TODO Auto-generated constructor stub

}

MemAccessMonitor::~MemAccessMonitor() {
	// TODO Auto-generated destructor stub
}
unsigned long timestamp = 0;
void MemAccessMonitor::markAccess(map<char *, ThreadMemInfo*> *memoryAccess,
		char* argName, PTR_TYPE address, char *threadLogicalName) {
	PTR_TYPE memLineAddress = getMemLineAddressByMemAddress(address);
	int cacheEntry = (memLineAddress >> MEM_MASK_SIZE_IN_BITS)
			% CACHE_SIZE_IN_LINES;
	//	SET THE START TIMESTAMP IF NEEDED
	//	SET THE END TIMESTAMP ANYWAY
	//printf("mem line address 0x%x accessing cash entry %d \n",memLineAddress,cacheEntry);
	pthread_mutex_lock( &mutex1 );
	ThreadMemInfo* tmi = memoryAccess[cacheEntry][threadLogicalName];
	if (tmi == NULL) {
		tmi = memoryAccess[cacheEntry][threadLogicalName] = new ThreadMemInfo();
		tmi->threadStartTimestamp = timestamp;
		tmi->cacheLineNumber = cacheEntry;
	}
	pthread_mutex_unlock( &mutex1 );
	tmi->threadEndTimestamp = timestamp;
	tmi->memAccessCount++;
	timestamp++;
}

void MemAccessMonitor::printCacheAccess() {
#ifdef DEBUG
	printf("\nprinting all read access:\n");

	for (int i = 0; i < CACHE_SIZE_IN_LINES; i++) {
		map<char *, ThreadMemInfo*> memReadsForLineI = memoryReads[i];
		if (memReadsForLineI.size() > 0) {
			printf("printing mem access for line: %d\n", i);
			map<char *, ThreadMemInfo*>::iterator itr;
			for (itr = memReadsForLineI.begin(); itr != memReadsForLineI.end(); itr++) {
				ThreadMemInfo* tmi = itr->second;
				printf("thread %s : accessed address %d ; start %d ; end %d\n",
						itr->first, tmi->memAccessCount,
						tmi->threadStartTimestamp, tmi->threadEndTimestamp);
			}
		}
	}
	printf("\nprinting all Write access:\n");
	for (int i = 0; i < CACHE_SIZE_IN_LINES; i++) {
		map<char *, ThreadMemInfo*> memReadsForLineI = memoryWrites[i];
		if (memReadsForLineI.size() > 0) {
			printf("printing mem access for line: %d\n", i);
			map<char *, ThreadMemInfo*>::iterator itr;
			for (itr = memReadsForLineI.begin(); itr != memReadsForLineI.end(); itr++) {
				ThreadMemInfo* tmi = itr->second;
				printf("thread %s : accessed address %d ; start %d ; end %d\n",
						itr->first, tmi->memAccessCount,
						tmi->threadStartTimestamp, tmi->threadEndTimestamp);
			}
		}
	}
#endif
}

void MemAccessMonitor::printScheduleAdvice() {
	map<char*, ThreadMemInfo *> writingThreads;
	// go over all write access and "remember" the threads involved in writing;
	// also; remember their start time and end time, and sum it to create minimum start and maximum end.
	for (int i = 0; i < CACHE_SIZE_IN_LINES; i++) {
		map<char *, ThreadMemInfo*> memReadsForLineI = memoryWrites[i];
		if (memReadsForLineI.size() > 0) {
			map<char *, ThreadMemInfo*>::iterator itr;
			for (itr = memReadsForLineI.begin(); itr != memReadsForLineI.end(); itr++) {
				// save the minimal start time of the thread and it's maximal end time
				// also save the maximal accessed memory address (with it's access count)
				if (writingThreads[itr->first] == NULL) {
					writingThreads[itr->first] = new ThreadMemInfo();
					writingThreads[itr->first]->threadEndTimestamp
							= itr->second->threadEndTimestamp;
					writingThreads[itr->first]->threadStartTimestamp
							= itr->second->threadStartTimestamp;
					writingThreads[itr->first]->cacheLineNumber
							= itr->second->cacheLineNumber;
					writingThreads[itr->first]->memAccessCount
							= itr->second->memAccessCount;
				} else {
					if (writingThreads[itr->first]->threadEndTimestamp
							< itr->second->threadEndTimestamp)
						writingThreads[itr->first]->threadEndTimestamp
								= itr->second->threadEndTimestamp;
					if (writingThreads[itr->first]->threadStartTimestamp
							> itr->second->threadStartTimestamp)
						writingThreads[itr->first]->threadStartTimestamp
								= itr->second->threadStartTimestamp;
					if (writingThreads[itr->first]->memAccessCount
							< itr->second->memAccessCount) {
						writingThreads[itr->first]->memAccessCount
								= itr->second->memAccessCount;
						writingThreads[itr->first]->cacheLineNumber
								= itr->second->cacheLineNumber;
					}
				}

			}
		}
	}
	// if there are more (or equal) processors than threads; return
	if (writingThreads.size() <= procnum)
		return;
	unsigned int availProcessorsNum = procnum;
	list<char*> processors[availProcessorsNum];

	map<char*, ThreadMemInfo *>::iterator threadToSchedItr;
	//	iterate on all the threads
	map<char*, char*> alreadyScheduled;
	for (threadToSchedItr = writingThreads.begin(); threadToSchedItr
			!= writingThreads.end(); threadToSchedItr++) {
		if (alreadyScheduled[threadToSchedItr->first])
			continue;
#ifdef DEBUG
		printf("thread %s start time %d and end time %d\n",
				threadToSchedItr->first,
				threadToSchedItr->second->threadStartTimestamp,
				threadToSchedItr->second->threadEndTimestamp);
#endif

		//find if there are more than 4 threads running at the same time:
		list<char*> sameTimeThreads;
		sameTimeThreads.push_back(threadToSchedItr->first);
#ifdef DEBUG
		printf("threads that run at the same time of %s\n",
				threadToSchedItr->first);
#endif
		map<char*, ThreadMemInfo *>::iterator innerItr;
		for (innerItr = writingThreads.begin(); innerItr
				!= writingThreads.end(); innerItr++) {
			// if not looking on the same thread
			if (alreadyScheduled[innerItr->first])
				continue;
			// if not using the same cache line as their prime writing address
			if (innerItr->second->cacheLineNumber
					!= threadToSchedItr->second->cacheLineNumber)
				continue;
			if (strcmp(innerItr->first, threadToSchedItr->first)) {
				//if both run at the same time (if started before and finished after|| if started after and didnt finish before
				if ((threadToSchedItr->second->threadStartTimestamp
						< innerItr->second->threadStartTimestamp
						&& threadToSchedItr->second->threadEndTimestamp
								> innerItr->second->threadStartTimestamp)
						|| (innerItr->second->threadStartTimestamp
								< threadToSchedItr->second->threadStartTimestamp
								&& innerItr->second->threadEndTimestamp
										> threadToSchedItr->second->threadStartTimestamp)) {

					sameTimeThreads.push_back(innerItr->first);
				}
			}
		}
		//		if there are not enough processors that run at the same time; look to schedule:
		if (sameTimeThreads.size() > availProcessorsNum) {
#if DEBUG
			printf(" %d > %d ",sameTimeThreads.size() , availProcessorsNum);
#endif
			int numberOfThreadsToClusterEachTime =sameTimeThreads.size() / availProcessorsNum ;
			int numberOfThreadsToCluster =
								sameTimeThreads.size() / availProcessorsNum + sameTimeThreads.size() % availProcessorsNum;
			// randomize:
			int generalIterator= rand();
			for (unsigned int processorsItr = 0; processorsItr
					< availProcessorsNum; processorsItr++) {

				list<char*> * clustered = maximalAccess(
						numberOfThreadsToCluster, sameTimeThreads,
						writingThreads, alreadyScheduled);
				processors[generalIterator++ % procnum].merge(*clustered);
				delete clustered;
				numberOfThreadsToCluster=numberOfThreadsToClusterEachTime;
			}

		}
	}
	printToFile(processors);
}

#define FILENAME "..\\threads\\scheduler.cpp"
void MemAccessMonitor::printToFile(list<char*> *processors){
	printf("Started printing to file- %s\n",FILENAME);

	ofstream scheduler(FILENAME, ios::out);
	scheduler<<"#include <map>\n#include <string.h>\nusing namespace std;\n\n";
	scheduler<<"struct cmp_str\n{\n bool operator()(char const *a, char const *b)\n";
	scheduler<<" {\n  return strcmp(a, b) < 0;\n }\n";
	scheduler<<"};\n\n";
	scheduler<<"map<char*,int,cmp_str> threadToProcessor;\n\n\n";
	scheduler<<"void loadSchedule(){\n";
	list<char*>::iterator itr;
	for (int i=0; i<procnum; i++){
		for (itr = processors[i].begin(); itr
			!= processors[i].end(); itr++) {
#if DEBUG
			printf("processor %d has %s allocated to it ... \n",i,*itr);
#endif
			// never put zero so , because it's the default value where there is no value (it's the null value);
			scheduler<<"	threadToProcessor[\""<<*itr<<"\"]="<<(i+1)<<";\n";
			cout<<"	threadToProcessor[\""<<*itr<<"\"]="<<(i+1)<<";\n";
		}
	}
	scheduler<<"}\n";
	scheduler.close();
	printf("Ended printing to file- %s\n",FILENAME);
}

list<char*> *MemAccessMonitor::maximalAccess(int groupSize,
		list<char*> &sameTimeThreads,
		map<char*, ThreadMemInfo *> &writingThreads,
		map<char*, char*> &alreadyScheduled) {
	static int clusterNum=0;
	list<char*> * result = new list<char*> ();
	list<char*>::iterator sameTimeItr;
	char* maxTimeThreadName;
	int accessTimes = 0;
	int cnum=clusterNum++;
#ifdef DEBUG
	printf("clustering threads on a processor - Cluster num %d group size: %d \n",cnum,groupSize);
#endif
	for (int i = 0; i < groupSize && sameTimeThreads.size() > 0; i++) {
		//find maximal access time
		for (sameTimeItr = sameTimeThreads.begin(); sameTimeItr
				!= sameTimeThreads.end(); sameTimeItr++) {
			if (accessTimes <= writingThreads[*sameTimeItr]->memAccessCount) {
				accessTimes = writingThreads[*sameTimeItr]->memAccessCount;
				maxTimeThreadName = *sameTimeItr;
			}
		}
		sameTimeThreads.remove(maxTimeThreadName);
		result->push_back(maxTimeThreadName);
		alreadyScheduled[maxTimeThreadName] = maxTimeThreadName;
		accessTimes = 0;
#ifdef DEBUG
		printf("%s clustered the processor Cluster Num %d\n", maxTimeThreadName,cnum);
#endif

	}
	return result;
}
void MemAccessMonitor::markWrite(char* argName, PTR_TYPE address,
		char *threadLogicalName) {

	if (threadLogicalName) {
		//		printf("Writing %s at address 0x%x (for thread %s)\n",argName,address,threadLogicalName);
		markAccess(memoryWrites, argName, address, threadLogicalName);
	}
}

void MemAccessMonitor::markRead(char* argName, PTR_TYPE address,
		char *threadLogicalName) {
	//	printf("Reading %s at address 0x%x (for thread %s)\n",argName,address,threadLogicalName);

	if (threadLogicalName)
		markAccess(memoryReads, argName, address, threadLogicalName);
}

MemAccessMonitor *MemAccessMonitor::instance;

MemAccessMonitor *MemAccessMonitor::getInstance() {
	if (instance == NULL) {
		instance = new MemAccessMonitor();
	}
	return instance;
}
