#define PROC_NUM 8
/*Minimal cache line size should be 2, in order to hold both res[x][0] and res[x][1]*/
/*size of a cache line in core i is 64 bytes*/
/*size of long is 6 bytes*/
#define CACHE_LINE_SIZE 4

int main(int argc, char **argv) {
	int REPETITION = 100;
	long STOP_ITER = 1000000;
	long ITER = STOP_ITER * 2;
	int i, firstProc, secondProc;
	int repetition;
	long relativeResults[PROC_NUM][PROC_NUM];
	/*res[x][0] is used to hold the progress of a core during test*/
	/*res[x][1] is used as a flag to indicate that a core has reached the STOP_ITER */
	long res[PROC_NUM][CACHE_LINE_SIZE];
	long summedRes[PROC_NUM][CACHE_LINE_SIZE];
	long relativeResultsAVG[PROC_NUM][PROC_NUM];
	long relativeResultsMAX[PROC_NUM][PROC_NUM];
	/*first parameter should be repetition number for average results*/
	if (argc > 1) {
		REPETITION = atoi(argv[1]);
		/* second parameter is the number of iterations to measure relative speed */
		if (argc > 2) {
			STOP_ITER = atoi(argv[2]) * 1000000;
			ITER = STOP_ITER * 2;
		}
	}
	/*initialize with a parfor in order to load threads mechanism and remove initializations overheads from
	 the real relative rates calculations*/
	parfor (i = 0; i < PROC_NUM; i++) {
				res[i][0] = 0;
				summedRes[i][0] = 0;
				res[i][1] = 0;
			}
			/*initialize variables*/
			for (firstProc = 0; firstProc < PROC_NUM; firstProc++) {
				for (secondProc = 0; secondProc < PROC_NUM; secondProc++) {
					relativeResults[firstProc][secondProc] = 0;
					relativeResultsAVG[firstProc][secondProc] = 0;
					relativeResultsMAX[firstProc][secondProc] = 0;
				}
			}
			for (repetition = 0; repetition < REPETITION; repetition++) {
				//printf("Round number %d\n",repetition+1);
				for (firstProc = 0; firstProc < PROC_NUM; firstProc++) {
					for (secondProc = firstProc + 1; secondProc < PROC_NUM; secondProc++) {
						res[firstProc][0] = 0;
						res[secondProc][0] = 0;
						res[firstProc][1] = 0;
						res[secondProc][1] = 0;
						//printf("run processor %d against %d\n",firstProc,secondProc);
						parfor (i = 0; i < PROC_NUM; i++) {
									long j;
									int k;
									if (i == firstProc || i == secondProc) {
										for (j = 0; j < ITER; j++) {
											res[i][0]++;
											if (res[i][0] >= STOP_ITER
													&& res[i][1] == 0) {
												if (i == firstProc) {
													relativeResults[firstProc][secondProc]
															= res[secondProc][0]
																	- STOP_ITER;
													relativeResultsAVG[firstProc][secondProc]
															+= res[secondProc][0]
																	- STOP_ITER;

													if (abs(
															relativeResultsMAX[firstProc][secondProc])
															<= abs(
																	relativeResults[firstProc][secondProc])) {
														relativeResultsMAX[firstProc][secondProc]
																= relativeResults[firstProc][secondProc];
													}

												} else {
													relativeResults[secondProc][firstProc]
															= res[firstProc][0]
																	- STOP_ITER;
													relativeResultsAVG[secondProc][firstProc]
															+= res[firstProc][0]
																	- STOP_ITER;
													if (abs(
															relativeResultsMAX[secondProc][firstProc])
															<= abs(
																	relativeResults[secondProc][firstProc])) {
														relativeResultsMAX[secondProc][firstProc]
																= relativeResults[secondProc][firstProc];
													}

												}

												res[i][1] = 1;
											}
										}
									}
								}
							}
						}
					}

					printf(
							"\n\nRelative Rates \n%d processors %d iterations %d byte spacing between the local variables of each thread\n",
							PROC_NUM, STOP_ITER, CACHE_LINE_SIZE * sizeof(long));
					printf("%d repetition for avarage and maximum rates \n",
							REPETITION);

					printf(
							"\\begin{table}\n\\begin{tabular}{c|cccc}\n\\hline\\hline\n");

					printf("    ");
					for (firstProc = 0; firstProc < PROC_NUM; firstProc++) {
						printf(" &  %3d", firstProc + 1);
					}
					printf(" \\\\ \\hline \n");

					for (firstProc = 0; firstProc < PROC_NUM; firstProc++) {
						printf(" %3d ", firstProc + 1);
						for (secondProc = 0; secondProc < PROC_NUM; secondProc++) {
							printf(
									"& %4d ",
									(int) (100
											* (((double) relativeResultsAVG[firstProc][secondProc])
													/ REPETITION) / STOP_ITER));
						}
						printf("\\\\\n");
					}
					printf(
							"\\hline\\hline\n\\end{tabular}\n\\caption[]{\\label{tab:rate}\\sl\n");
					for (i = 1; i < argc; i++)
						printf("%s ", argv[i]);
					printf("\n");
					printf("Avarage Rates\n");
					printf("\\end{table}\n\%\n");

					printf(
							"\\begin{table}\n\\begin{tabular}{c|cccc}\n\\hline\\hline\n");

					printf("    ");
					for (firstProc = 0; firstProc < PROC_NUM; firstProc++) {
						printf(" &  %3d", firstProc + 1);
					}
					printf(" \\\\ \\hline \n");

					for (firstProc = 0; firstProc < PROC_NUM; firstProc++) {
						printf(" %3d ", firstProc + 1);
						for (secondProc = 0; secondProc < PROC_NUM; secondProc++) {
							printf(
									"& %4d ",
									(int) (100
											* ((double) relativeResultsMAX[firstProc][secondProc])
											/ STOP_ITER));
						}
						printf("\\\\\n");
					}

					printf(
							"\\hline\\hline\n\\end{tabular}\n\\caption[]{\\label{tab:rate}\\sl\n");
					for (i = 1; i < argc; i++)
						printf("%s ", argv[i]);
					printf("\n");
					printf("Maximal Rates\n");
					printf("\\end{table}\n\%\n");

				}
