/*
 * ParCPPPrintTraversal.cpp
 *
 *  Created on: 08/02/2009
 *      Author: Gil Kulish
 */

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

 CTool Library
 Copyright (C) 1998-2001	Shaun Flisakowski

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 1, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
 o+
 o+     Programmer:   Stefan Seefeld
 o+     Date:         2003-07-31
 o+
 o+     Print an AST
 o+
 o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

/**
 * Gil Kulish:
 * this file was copy-pasted from PrintTraversal.cpp
 * changes are done in the parfor method.
 * a major change was the condition "if (isPrinting)"
 * added before each and every print, it's purpose is to disable printing of the code
 * unless inside a parfor/parblock (because all other code is printed into another file anyway by PrintTraversal),
 * when inside the PARFOR/PARBLOCK there is some parsing and processing, !but! the main block of the PARFOR/PARBLOCK
 * is passed to PrintTraversal for further printing and processing.
 */
#include <ctool/decl.h>
#include <ctool/stemnt.h>
#include <ctool/project.h>
#include <ctool/express.h>
#include <ctool/symbol.h>
#include <iostream>
#include <sstream>
#include <set>
#include <string>
#include <cstring>
#include <ctool/PrintTraversal.h>
#include <ctool/ParCPPPrintTraversal.h>
#include <ctool/ParVarListTraversal.h>
using namespace std;
//static char *openRemark=" /* ";
//static char *closeRemark=" */ ";
//static char *closeOpenRemark=" *//* ";

string trim(string& str);
string replace(string orig, string whatToReplace, string replaceWith);

ParCPPPrintTraversal::ParCPPPrintTraversal(std::ostream &os, bool d,
		std::ostream &headerOut) :
	PrintTraversal(os, d, headerOut), isPrinting(false) {
}

void ParCPPPrintTraversal::traverse_base(BaseType *type) {
}

void ParCPPPrintTraversal::traverse_ptr(PtrType *ptr) {
}

void ParCPPPrintTraversal::traverse_array(ArrayType *) {
}

void ParCPPPrintTraversal::traverse_bit_field(BitFieldType *) {
}

void ParCPPPrintTraversal::traverse_function(FunctionType *) {
}

void ParCPPPrintTraversal::traverse_symbol(Symbol *node) {
	if (isPrinting) {
		out << *node;

	}
}

void ParCPPPrintTraversal::traverse_int(IntConstant *node) {
	if (isPrinting)
		out << node->lng;
	if (node->L)
		if (isPrinting)
			out << "L";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
		 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 		out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_uint(UIntConstant *node) {
	if (isPrinting)
		out << node->ulng;
	if (node->L)
		if (isPrinting)
			out << "L";

	/*#ifdef    SHOW_TYPES
	 if (node->type != NULL)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_float(FloatConstant *node) {
	if (isPrinting)
		out << node->value;

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_char(CharConstant *node) {
	if (node->wide)
		if (isPrinting)
			out << 'L';

	if (isPrinting)
		out << "'";
	meta_char(node->ch, false);
	if (isPrinting)
		out << "'";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_string(StringConstant *node) {
	if (node->wide)
		if (isPrinting)
			out << 'L';

	if (isPrinting)
		out << '"';
	meta_string(node->buff);
	if (isPrinting)
		out << '"';

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_array(ArrayConstant *node) {
	if (isPrinting)
		out << "{ ";

	if (!node->items.empty()) {
		ExprVector::const_iterator i = node->items.begin();
		(*i)->accept(this);
		for (++i; i != node->items.end(); ++i) {
			if (isPrinting)
				out << ", ";
			(*i)->accept(this);
		}
	}

	if (isPrinting)
		out << " }";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_enum(EnumConstant *node) {
	traverse_symbol(node->name);
	if (node->value) {
		if (isPrinting)
			out << " = ";
		node->value->accept(this);
	}

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_variable(Variable *node) {
	traverse_symbol(node->name);

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_call(FunctionCall *node) {
	if (node->function->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->function->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->function->accept(this);

	if (isPrinting)
		out << "(";

	if (!node->args.empty()) {
		ExprVector::const_iterator i = node->args.begin();
		if (((*i)->etype == ET_BinaryExpr)
				&& ((static_cast<BinaryExpr*> (*i))->op() == BO_Comma)) {
			if (isPrinting)
				out << "(";
			(*i)->accept(this);
			if (isPrinting)
				out << ")";
		} else
			(*i)->accept(this);

		for (++i; i != node->args.end(); ++i) {
			if (isPrinting)
				out << ",";

			if (((*i)->etype == ET_BinaryExpr)
					&& ((static_cast<BinaryExpr*> (*i))->op() == BO_Comma)) {
				if (isPrinting)
					out << "(";
				(*i)->accept(this);
				if (isPrinting)
					out << ")";
			} else
				(*i)->accept(this);
		}
	}

	if (isPrinting)
		out << ")";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_unary(UnaryExpr *node) {
	if (isPrinting)
		node->print(out);
}

void ParCPPPrintTraversal::traverse_binary(BinaryExpr *node) {
	if (isPrinting)
		node->print(out);
}

void ParCPPPrintTraversal::traverse_trinary(TrinaryExpr *node) {
	if (isPrinting)
		node->print(out);
}

void ParCPPPrintTraversal::traverse_assign(AssignExpr *node) {
	if (node->lValue()->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->lValue()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->lValue()->accept(this);

	if (isPrinting)
		out << " ";
	assign_op(node->aOp);
	if (isPrinting)
		out << " ";

	if (node->rValue()->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->rValue()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->rValue()->accept(this);

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_rel(RelExpr *node) {
	if (node->leftExpr()->precedence() <= node->precedence()) {
		if (isPrinting)
			out << "(";
		node->leftExpr()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->leftExpr()->accept(this);

	if (isPrinting)
		out << " ";
	rel_op(node->rOp);
	if (isPrinting)
		out << " ";

	if (node->rightExpr()->precedence() <= node->precedence()) {
		if (isPrinting)
			out << "(";
		node->rightExpr()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->rightExpr()->accept(this);

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_cast(CastExpr *node) {

	if (isPrinting){
		out << "(";
	node->castTo->printType(out, NULL, true, 0);
	}
	if (isPrinting)
		out << ") ";

	if (isPrinting){
		out << "(";
	node->expr->accept(this);
		out << ")";
	}
	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/

}

void ParCPPPrintTraversal::traverse_sizeof(SizeofExpr *node) {
	if (isPrinting)
		out << "sizeof(";
	if (node->sizeofType)
		node->sizeofType->printType(out, NULL, true, 0);
	else
		node->expr->accept(this);
	if (isPrinting)
		out << ") ";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << " closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_index(IndexExpr *node) {
	if (node->array->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->array->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->array->accept(this);

	if (isPrinting)
		out << "[";
	node->_subscript->accept(this);
	if (isPrinting)
		out << "]";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParCPPPrintTraversal::traverse_label(Label *node) {
	decr_indent();
	indent();
	incr_indent();

	switch (node->type) {
	case LT_None:
		assert(0);
		break;

	case LT_Default:
		if (isPrinting)
			out << "default";
		break;

	case LT_Case:
		assert(node->begin);
		if (isPrinting)
			out << "case ";
		node->begin->accept(this);
		break;

	case LT_CaseRange:
		assert(node->begin);
		assert(node->end);
		if (isPrinting)
			out << "case ";
		node->begin->accept(this);
		if (isPrinting)
			out << " ... ";
		node->end->accept(this);
		break;

	case LT_Goto:
		assert(node->name);
		traverse_symbol(node->name);
		break;
	}
	if (isPrinting)
		out << ":\n";
}

void ParCPPPrintTraversal::traverse_decl(Decl *node) {
	if (isPrinting)
		node->print(out, show_base, level);
}

void ParCPPPrintTraversal::traverse_statement(Statement *node) {
	if (isPrinting)
		node->print(out, level);
}

void ParCPPPrintTraversal::traverse_file_line(FileLineStemnt *node) {
	/*#ifdef PRINT_LOCATION
	 if (isPrinting)
	 out << "$<FileLineStemnt:";
	 node->location.printLocation(out);
	 if (isPrinting)
	 out << ">$";
	 #endif*/
	if (node->linenumber > 0)
		if (isPrinting)
			out << "#line " << node->linenumber;
		else if (isPrinting)
			out << "#file";
	if (isPrinting)
		out << " \"" << node->filename << "\"" << std::endl;
}

void ParCPPPrintTraversal::traverse_include(InclStemnt *node) {
	/*#ifdef PRINT_LOCATION
	 if (isPrinting)
	 out << "$<InclStemnt:";
	 node->location.printLocation(out);
	 if (isPrinting)
	 out << ">$";
	 #endif*/
	/* Gil Kulish - don't use isPrinting here, print include files*/
	//	if (isPrinting){
	out << "#include " << (node->isStandard ? '<' : '"');
	out << node->filename << (node->isStandard ? '>' : '"') << std::endl;
	//}
}

void ParCPPPrintTraversal::traverse_end_include(EndInclStemnt *node) {
	/*#ifdef PRINT_LOCATION
	 if (isPrinting)
	 out << "$<EndInclStemnt:";
	 node->location.printLocation(out);
	 if (isPrinting)
	 out << ">$";
	 #endif*/
}

void ParCPPPrintTraversal::traverse_expression(ExpressionStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* ExpressionStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	assert(node->expression);
	node->expression->accept(this);
}

void ParCPPPrintTraversal::traverse_if(IfStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* IfStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();

	if (isPrinting)
		out << "if (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")\n";
	block(node->thenBlk);

	if (node->elseBlk) {
		if (isPrinting)
			out << std::endl;
		indent();
		if (isPrinting)
			out << "else\n";
		block(node->elseBlk);
	}
}

void ParCPPPrintTraversal::traverse_switch(SwitchStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* SwitchStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "switch (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
}

void ParCPPPrintTraversal::traverse_for(ForStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* ForStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "for (";
	if (node->init)
		node->init->accept(this);
	if (isPrinting)
		out << "; ";

	if (node->cond)
		node->cond->accept(this);
	if (isPrinting)
		out << "; ";

	if (node->incr)
		node->incr->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
}

/* Gil Kulish added: */
void ParCPPPrintTraversal::traverse_for(ParForStemnt *node) {
	parforsDeapth++;
	parforName[parforsDeapth]++;
	/**get the variables list that are used inside the parfor and are
	 declared outside the parfor, that is- variables that are global to the
	 parfor. (from the parfor point of view)*/
	ostringstream parforCtrSig;
	ostringstream printIndexS;

	ParVarListTraversal initInParFor(out, debug, level + 3);
	//use printIndex to find the name of the iterator inside the parfor loop
	ParVarListTraversal printIndex(printIndexS, debug, level + 3);
	ParVarListTraversal pvl(parforCtrSig, debug, level + 3);

	std::string iterName;
	if (node->init) {
		AssignExpr* ae = ((AssignExpr*) node->init);
		printIndex.isPrinting = true;
		ae->leftExpr()->accept(&printIndex);
		iterName = printIndexS.str();
	}

	if (node->init)
		node->init->accept(&initInParFor);

	if (node->cond)
		node->cond->accept(&initInParFor);

	if (node->incr)
		node->incr->accept(&initInParFor);

	node->accept(&pvl);

	incr_indent();
	isPrinting = true;
	if (debug) {
		if (isPrinting)
			out << "/* ParForStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	if (isPrinting) {
		out << "\nclass ParForJob" << getParforName()
				<< " : public ParForJob {\n";
	}
	string insideCtr("");
	string insideCallToCtr("");
	indent();
	out << "public:\n\n";
	indent();
	out << "static long threadNumOutOfGroupSequance;\n\n";
	indent();
	out << "long threadNumOutOfGroup;\n\n";
	indent();

	out << "virtual void run(){\n";
	/** Gil Kulish - recursivally create all the rest of the Parfor\parblock calls */
	incr_indent();
	PrintTraversal pt(out, debug, headerOut);
	pt.level = level-2;
	pt.parforsDeapth = parforsDeapth; // increment the depth
	pt.iterName = iterName;
	pt.parforOuterLoop = node;
	indent();
	out<<"threadNumToLogicalName[(PTR_TYPE)(getPthThreadLocal())]=threadLogicalName;\n";

	for (int i=0; i<MAX_DEAPTH; i++)
		PrintTraversal::parforName[i]=ParCPPPrintTraversal::parforName[i];
	RelExpr *ae = (RelExpr*) node->cond;
//	RelExpr *initRight = (RelExpr*) node->init; <- not needed at this point

	indent();
	out << "for (" << iterName << "=__start;" << iterName;
	// gil kulish todo: only supports i < 5, support also 4<i
	rel_op(ae->rOp);
	out << "__end;";
	if (node->incr)
		node->incr->accept(this);
	out << ")";
	pt.block(node->block);
	indent();
	out << "signalFinished();\n";
	decr_indent();
	indent();
	out << "}\n\n";

	set<Decl*> varInitList = *(initInParFor.getVarList());
	set<Decl*> varList = *(pvl.getVarList());
	// make sure that the "i" in for (i=5; i<10; i++) cannot be changed
	// inside the loop

	for (set<Decl*>::iterator i = varList.begin(); i != varList.end(); i++) {
		char qual = '&';
		indent();
		// check if the variable is inside the for barracks (for (i=10; i<100; i++)
		if (varInitList.find(*i) != varInitList.end()) {
			//out<<"const "; //Gil Kulish (temporary disabled)
			//parforCtrSig<<"const "; //Gil Kulish (temporary disabled)
			qual = ' ';
		}
		initInParFor.printDecl(*i, qual);
		pvl.printDecl(*i, qual);
		out << ";\n\n";
		parforCtrSig << ",";
		insideCtr += (*i)->name->name + "(" + (*i)->name->name + "),";
		insideCallToCtr += (*i)->name->name + ",";
	}
	insideCallToCtr += "__start,__end";
	parforCtrSig << "int __start,int __end";
	insideCtr += "ParForJob(__start,__end)";
	//	indent();
	//	out<<"int __end;\n\n"; <- this went into ParForJob.cpp
	indent();
	string sparforCtrSig(parforCtrSig.str());
	sparforCtrSig = trim(sparforCtrSig);
/* generate the constructor of the thread */
	out << "ParForJob" << getParforName() << "("
			<< sparforCtrSig << "):\n";
	indent();
	indent();
	out << insideCtr << "{\n";
	indent();
	out<<"	threadNumOutOfGroup=threadNumOutOfGroupSequance++;\n";
	indent();
	out<<"	sprintf(threadLogicalName,\"%s%ld\",threadGroupName,threadNumOutOfGroup);\n";
	indent();
	out<<"}\n\n";

	/* Gil Kulish add support for lazy thread:*/
	indent();
	out << "void create_lazy_thread(int __start, int __end){\n";
	incr_indent();
	indent();
	out << "static int processorsIter=0;\n";
	indent();
	out << "if (!lazyThread){\n";
	incr_indent();
	indent();
	out << "lazyThread=new UserThread(new ParForJob";
	out << getParforName()<< "(" << insideCallToCtr
			<< "));\n";
	indent();
	out << "lazyThread->setPJ(pj);\n";
	indent();
	out << "lazyThread->parentThread=this;\n";
	indent();
	out << "lazyThread->toRun->parentThread=this;\n";
	indent();
	out << "pj->signalNewThreadOpened();\n";
	indent();
	out << "ThreadLocalData *tld = getPthThreadLocal();\n";
	indent();
	out << "ProcessorsManager *pm=ProcessorsManager::getInstance();\n";
	indent();
	out << "pm->execute(lazyThread);\n";
	decr_indent();
	indent();
	out << "}\n";
	decr_indent();
	indent();
	out << "}\n\n";
	indent();
	out << "static char* threadGroupName;\n\n";
	indent();
	out << "char *getThreadGroupName(){\n";
	incr_indent();
	indent();
	out << "return threadGroupName;\n";
	decr_indent();
	indent();
	out << "}\n\n";

	decr_indent();
	decr_indent();
	/* End Gil Kulish add support for lazy thread:*/
	out << "};\n\n";
	indent();
	out << "char* ParForJob" << getParforName()
			<< "::threadGroupName=\"" << getParforName()
			<< "\";\n\n";
	indent();
	out<<"long ParForJob" << getParforName()<<"::threadNumOutOfGroupSequance=0;\n\n";

	if (parforsDeapth <= 0) {// if deapth >0 it means we're not calling this from the _par.cpp file
		headerOut << "#ifdef __cplusplus\n";
		headerOut << "   extern \"C\" void parfor" << getParforName()<< "(" << sparforCtrSig << ");\n";
		headerOut << "#else\n";
		headerOut << "   void parfor" << getParforName()
				<< "(" << replace(sparforCtrSig, "&", "*") + ");\n";
		headerOut << "#endif\n";

	} else {
		headerOut << "#ifdef __cplusplus\n";
		headerOut << "    void parfor" << getParforName()
				<< "(" << sparforCtrSig + ");\n";
		headerOut << "#endif\n";
	}

	if (isPrinting) {
		out << "void parfor" << getParforName()<< "("
				<< sparforCtrSig << "){\n\n";
		incr_indent();
		indent();
		out << "int origStart=__start;\n";
		indent();
		out << "int origEnd=__end;\n";
		indent();
		out << "ParJoin pj(0);\n";
		indent();
		out << "Thread** threads=new Thread*[procnum];\n";
		indent();
		out << "int processorsIter=0;\n";
		indent();
		out << "ProcessorsManager *pm=ProcessorsManager::getInstance();\n";
		indent();
		out
				<< "for (processorsIter=0; processorsIter<procnum;processorsIter++){\n";
		indent();
		out
				<< "	__start=(int)(origStart+(origEnd-origStart)*((double)processorsIter/procnum));\n";
		indent();
		out
				<< "	__end=(int)(origStart+(origEnd-origStart)*((processorsIter+1.)/procnum));\n";
		indent();
		out << "	threads[processorsIter]=new Thread(new ParForJob";
		out << getParforName() << "(" << insideCallToCtr
				<< "));\n";
		indent();
		out << "	pj.signalNewThreadOpened();\n";
		indent();
		out << "	threads[processorsIter]->setPJ(&pj);\n";
		indent();
		out << "	pm->execute(threads[processorsIter]);\n";
		indent();
		out << "}\n";
		indent();
		out << "pj.join();\n";
		out << "}\n\n\n";
	}
	//	gil kulish - at this point there is no need for an increment  (we'll assume exact "x=n;x<m;x++")
	//	if (node->incr)
	//		node->incr->accept(this);
	//	if (isPrinting)
	//		out << ")\n";


	isPrinting = false;
	/*Gil Kulish: recursivally call ParCPPPrintTraversal; to handle inner parfor/parblock*/
	//keep the same numbering as this parfor, (depth is increased):
	ParCPPPrintTraversal pct(out, debug, headerOut);
	pct.level = level;
	pct.parforsDeapth = parforsDeapth; // set the deapth
	pct.block(node->block);
	parforName[parforsDeapth+1]='A'-1;
	parforsDeapth--;

}
/* end of Gil Kulish added: */

void ParCPPPrintTraversal::traverse_while(WhileStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* WhileStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "while (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
}

void ParCPPPrintTraversal::traverse_do_while(DoWhileStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* DoWhileStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "do ";

	if (!node->block->isBlock())
		if (isPrinting)
			out << std::endl;

	block(node->block);
	if (!node->block->isBlock())
		if (isPrinting)
			out << std::endl;

	indent();
	if (isPrinting)
		out << "while (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")";
}

void ParCPPPrintTraversal::traverse_goto(GotoStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* GotoStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "goto ";
	traverse_symbol(node->dest);
}

void ParCPPPrintTraversal::traverse_return(ReturnStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* ReturnStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "return";

	if (node->result) {
		if (isPrinting)
			out << ' ';
		node->result->accept(this);
	}
}

void ParCPPPrintTraversal::traverse_declaration(DeclStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* DeclStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (!node->decls.empty()) {
		DeclVector::const_iterator i = node->decls.begin();

		traverse_decl(*i);
		show_base = false;
		for (++i; i != node->decls.end(); ++i) {
			if (isPrinting)
				out << ", ";
			traverse_decl(*i);
		}
		show_base = true;
	}

	if (isPrinting)
		out << ";";
}

void ParCPPPrintTraversal::traverse_typedef(TypedefStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* TypedefStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();

	if (isPrinting)
		out << "typedef ";

	if (!node->decls.empty()) {
		DeclVector::const_iterator i = node->decls.begin();
		traverse_decl(*i);
		for (++i; i != node->decls.end(); ++i) {
			if (isPrinting)
				out << ", ";
			traverse_decl(*i);
		}
	}

	if (isPrinting)
		out << ";";
}

void ParCPPPrintTraversal::traverse_block(Block *node) {
	if (debug) {
		if (isPrinting)
			out << "/* BlockStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "{\n";

	bool isDecl = (node->head != 0) ? node->head->isDeclaration() : false;
	for (Statement *statement = node->head; statement; statement
			= statement->next) {
		if (isDecl && !statement->isDeclaration()) {
			isDecl = false;
			if (isPrinting)
				out << std::endl;
		}

		incr_indent();
		statement->accept(this);
		decr_indent();

		if (statement->needSemicolon())
			if (isPrinting)
				out << ";";
		if (isPrinting)
			out << std::endl;
	}
	indent();
	if (isPrinting)
		out << "}\n";
}

void ParCPPPrintTraversal::traverse_function_definition(FunctionDef *node) {
	if (debug) {
		if (isPrinting)
			out << "/* FunctionDef:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	traverse_decl(node->decl);
	if (isPrinting)
		out << std::endl;

	traverse_block(node);
}

void ParCPPPrintTraversal::traverse_unit(TransUnit *unit) {
	int inInclude = 0;
	for (Statement *statement = unit->head; statement; statement
			= statement->next) {
		if (inInclude > 0) {
			if (statement->isEndInclude())
				inInclude--;
			else if (statement->isInclude())
				inInclude++;
		} else {
			if (statement->isInclude())
				inInclude++;
			statement->accept(this);
			if (isPrinting)
				out << std::endl;
		}
	}
	out << std::endl;
}

void ParCPPPrintTraversal::meta_char(char c, bool in) {
	switch (c) {
	case '\'':
		if (in)
			if (isPrinting)
				out << "'";
			else if (isPrinting)
				out << "\\'";
		break;

	case '"':
		if (in)
			if (isPrinting)
				out << "\\\"";
			else if (isPrinting)
				out << "\"";
		break;

	case '\0':
		if (isPrinting)
			out << "\\0";
		break;
	case '\\':
		if (isPrinting)
			out << "\\\\";
		break;
	case '\n':
		if (isPrinting)
			out << "\\n";
		break;
	case '\t':
		if (isPrinting)
			out << "\\t";
		break;
	case '\r':
		if (isPrinting)
			out << "\\r";
		break;
	case '\f':
		if (isPrinting)
			out << "\\f";
		break;
	case '\b':
		if (isPrinting)
			out << "\\b";
		break;
	case '\v':
		if (isPrinting)
			out << "\\v";
		break;
	case '\a':
		if (isPrinting)
			out << "\\a";
		break;
	case ESC_VAL:
		if (isPrinting)
			out << "\\e";
		break;

	default:
		// Show low and high ascii as octal
		if ((c < ' ') || (c >= 127)) {
			char octbuf[8];
			sprintf(octbuf, "%03o", (unsigned char) c);
			if (isPrinting)
				out << "\\" << octbuf;
		} else if (isPrinting)
			out << c;
		break;
	}
}

void ParCPPPrintTraversal::meta_string(const std::string &str) {
	for (unsigned i = 0; i < str.size(); ++i)
		meta_char(str[i], true);
}

void ParCPPPrintTraversal::rel_op(RelOp op) {
	switch (op) {
	case RO_Equal:
		if (isPrinting)
			out << "==";
		break;

	case RO_NotEqual:
		if (isPrinting)
			out << "!=";
		break;

	case RO_Less:
		if (isPrinting)
			out << "<";
		break;

	case RO_LessEql:
		if (isPrinting)
			out << "<=";
		break;

	case RO_Grtr:
		if (isPrinting)
			out << ">";
		break;

	case RO_GrtrEql:
		if (isPrinting)
			out << ">=";
		break;
	}
}

void ParCPPPrintTraversal::binary_op(BinaryOp op) {
	switch (op) {
	case BO_Plus:
		if (isPrinting)
			out << "+";
		break;

	case BO_Minus:
		if (isPrinting)
			out << "-";
		break;

	case BO_Mult:
		if (isPrinting)
			out << "*";
		break;

	case BO_Div:
		if (isPrinting)
			out << "/";
		break;

	case BO_Mod:
		if (isPrinting)
			out << "%";
		break;

	case BO_Shl:
		if (isPrinting)
			out << "<<";
		break;

	case BO_Shr:
		if (isPrinting)
			out << ">>";
		break;

	case BO_BitAnd:
		if (isPrinting)
			out << "&";
		break;

	case BO_BitXor:
		if (isPrinting)
			out << "^";
		break;

	case BO_BitOr:
		if (isPrinting)
			out << "|";
		break;

	case BO_And:
		if (isPrinting)
			out << "&&";
		break;

	case BO_Or:
		if (isPrinting)
			out << "||";
		break;

	case BO_Comma:
		if (isPrinting)
			out << ",";
		break;

	case BO_Member:
		if (isPrinting)
			out << ".";
		break;

	case BO_PtrMember:
		if (isPrinting)
			out << "->";
		break;

	default:
		//  case BO_Index        // x[y]
	case BO_Assign:
	case BO_Rel:
		break;
	}
}

void ParCPPPrintTraversal::unary_op(UnaryOp op) {
	switch (op) {
	case UO_Plus:
		if (isPrinting)
			out << "+";
		break;

	case UO_Minus:
		if (isPrinting)
			out << "-";
		break;

	case UO_BitNot:
		if (isPrinting)
			out << "~";
		break;

	case UO_Not:
		if (isPrinting)
			out << "!";
		break;

	case UO_PreInc:
	case UO_PostInc:
		if (isPrinting)
			out << "++";
		break;

	case UO_PreDec:
	case UO_PostDec:
		if (isPrinting)
			out << "--";
		break;

	case UO_AddrOf:
		if (isPrinting)
			out << "&";
		break;

	case UO_Deref:
		if (isPrinting)
			out << "*";
		break;
	}
}

void ParCPPPrintTraversal::assign_op(AssignOp op) {
	switch (op) {
	case AO_Equal:
		if (isPrinting)
			out << "=";
		break;

	case AO_PlusEql:
		if (isPrinting)
			out << "+=";
		break;

	case AO_MinusEql:
		if (isPrinting)
			out << "-=";
		break;

	case AO_MultEql:
		if (isPrinting)
			out << "*=";
		break;

	case AO_DivEql:
		if (isPrinting)
			out << "/=";
		break;

	case AO_ModEql:
		if (isPrinting)
			out << "%=";
		break;

	case AO_ShlEql:
		if (isPrinting)
			out << "<<=";
		break;

	case AO_ShrEql:
		if (isPrinting)
			out << ">>=";
		break;

	case AO_BitAndEql:
		if (isPrinting)
			out << "&=";
		break;

	case AO_BitXorEql:
		if (isPrinting)
			out << "^=";
		break;

	case AO_BitOrEql:
		if (isPrinting)
			out << "|=";
		break;
	}
}

void ParCPPPrintTraversal::block(Statement *node) {
	if (!node) {
		incr_indent();
		null();
		decr_indent();
	} else if (node->isBlock())
		node->accept(this);
	else {
		incr_indent();
		node->accept(this);
		decr_indent();

		if (node->needSemicolon())
			if (isPrinting)
				out << ";";
	}
}

char ParCPPPrintTraversal::parforName[MAX_DEAPTH]={'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1,'A'-1};

char* ParCPPPrintTraversal::getParforName(){
	char *name=new char[MAX_DEAPTH];
	strncpy(name,parforName,parforsDeapth+1);
	name[parforsDeapth+1]='\0';
	return name;
}

extern string trim(string& str);

string trim(string& str) {
	string::size_type pos1 = str.find_first_not_of(' ');
	string::size_type pos2 = str.find_last_not_of(' ');
	return str.substr(pos1 == string::npos ? 0 : pos1,
			pos2 == string::npos ? str.length() - 1 : pos2 - pos1 + 1);
}

string replace(string orig, string whatToReplace, string replaceWith) {
	string string1 = orig;
	string::size_type position = string1.find(whatToReplace); // find first space

	//
	while (position != string::npos) {
		string1.replace(position, whatToReplace.length(), replaceWith);
		position = string1.find(whatToReplace, position + replaceWith.length());
	}

	return string1;
}
