/*
 * ParVarListTraversal.cpp
 *
 *  Created on: 08/02/2009
 *      Author: Gil Kulish
 */

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

 CTool Library
 Copyright (C) 1998-2001	Shaun Flisakowski

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 1, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
 o+
 o+     Programmer:   Stefan Seefeld
 o+     Date:         2003-07-31
 o+
 o+     Print an AST
 o+
 o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

/**
 * Gil Kulish:
 * this file was copy-pasted from PrintTraversal.cpp
 * changes are done in the parfor method.
 * a major change was the condition "if (isPrinting)"
 * added before each and every print, it's purpose is to disable printing of the code
 * unless inside a parfor/parblock (because all other code is printed into another file anyway by PrintTraversal)
 */
#include <ctool/decl.h>
#include <ctool/stemnt.h>
#include <ctool/project.h>
#include <ctool/express.h>
#include <ctool/symbol.h>
#include <iostream>
#include <sstream>

#include <ctool/ParVarListTraversal.h>
using namespace std;

//static char *openRemark=" /* ";
//static char *closeRemark=" */ ";
//static char *closeOpenRemark=" *//* "; */

ParVarListTraversal::ParVarListTraversal(std::ostream &os, bool d,
		int indentationLevelToSearchVariablesFrom) :
	out(os), debug(d), indentLevel(0), show_base(true), parforLevel(0),
			isPrinting(false),
			indentLevelToSearchVariablesFrom(
					indentationLevelToSearchVariablesFrom) {
}

void ParVarListTraversal::traverse_base(BaseType *type) {
}

void ParVarListTraversal::traverse_ptr(PtrType *ptr) {
}

void ParVarListTraversal::traverse_array(ArrayType *) {
}

void ParVarListTraversal::traverse_bit_field(BitFieldType *) {
}

void ParVarListTraversal::traverse_function(FunctionType *) {
}

/** Gil Kulish:
 this is the main change in this file,
 this method creates the variables list used inside the parfor\parblock, but
 defined outside the parfor\parblock*/
void ParVarListTraversal::traverse_symbol(Symbol *node) {
	if (node->entry) {
		if (node->entry->type == VarDeclEntry||node->entry->type == ParamDeclEntry) {
			if ((!node->entry||!node->entry->scope)||(node->entry->scope->level <= indentLevelToSearchVariablesFrom)) {
				if (node->entry->uVarDecl) {
					varList.insert(node->entry->uVarDecl);
				}
			}
		}
	}
	if (isPrinting)
		out << *node;
}
ParVarListTraversal* ParVarListTraversal::activeInstance=NULL;

 ParVarListTraversal* ParVarListTraversal::getInstance(){
	return activeInstance;
}

void ParVarListTraversal::traverse_int(IntConstant *node) {
	if (isPrinting)
		out << node->lng;
	if (node->L)
		if (isPrinting)
			out << "L";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 	out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_uint(UIntConstant *node) {
	if (isPrinting)
		out << node->ulng;
	if (node->L)
		if (isPrinting)
			out << "L";

	/*#ifdef    SHOW_TYPES
	 if (node->type != NULL)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_float(FloatConstant *node) {
	if (isPrinting)
		out << node->value;

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_char(CharConstant *node) {
	if (node->wide)
		if (isPrinting)
			out << 'L';

	if (isPrinting)
		out << "'";
	meta_char(node->ch, false);
	if (isPrinting)
		out << "'";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_string(StringConstant *node) {
	if (node->wide)
		if (isPrinting)
			out << 'L';

	if (isPrinting)
		out << '"';
	meta_string(node->buff);
	if (isPrinting)
		out << '"';

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_array(ArrayConstant *node) {
	if (isPrinting)
		out << "{ ";

	if (!node->items.empty()) {
		ExprVector::const_iterator i = node->items.begin();
		(*i)->accept(this);
		for (++i; i != node->items.end(); ++i) {
			if (isPrinting)
				out << ", ";
			(*i)->accept(this);
		}
	}

	if (isPrinting)
		out << " }";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_enum(EnumConstant *node) {
	traverse_symbol(node->name);
	if (node->value) {
		if (isPrinting)
			out << " = ";
		node->value->accept(this);
	}

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_variable(Variable *node) {
	traverse_symbol(node->name);

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_call(FunctionCall *node) {
	if (node->function->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->function->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->function->accept(this);

	if (isPrinting)
		out << "(";

	if (!node->args.empty()) {
		ExprVector::const_iterator i = node->args.begin();
		if (((*i)->etype == ET_BinaryExpr)
				&& ((static_cast<BinaryExpr*> (*i))->op() == BO_Comma)) {
			if (isPrinting)
				out << "(";
			(*i)->accept(this);
			if (isPrinting)
				out << ")";
		} else
			(*i)->accept(this);

		for (++i; i != node->args.end(); ++i) {
			if (isPrinting)
				out << ",";

			if (((*i)->etype == ET_BinaryExpr)
					&& ((static_cast<BinaryExpr*> (*i))->op() == BO_Comma)) {
				if (isPrinting)
					out << "(";
				(*i)->accept(this);
				if (isPrinting)
					out << ")";
			} else
				(*i)->accept(this);
		}
	}

	if (isPrinting)
		out << ")";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_unary(UnaryExpr *node) {
	if (isPrinting)
		node->print(out);

	ostringstream stam;
	node->print(stam);
}

void ParVarListTraversal::traverse_binary(BinaryExpr *node) {
	if (isPrinting)
		node->print(out);

	ostringstream stam;
	node->print(stam);
}

void ParVarListTraversal::traverse_trinary(TrinaryExpr *node) {
	if (isPrinting)
		node->print(out);

	ostringstream stam;
	node->print(stam);
}

void ParVarListTraversal::traverse_assign(AssignExpr *node) {
	if (node->lValue()->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->lValue()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->lValue()->accept(this);

	if (isPrinting)
		out << " ";
	assign_op(node->aOp);
	if (isPrinting)
		out << " ";

	if (node->rValue()->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->rValue()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->rValue()->accept(this);

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_rel(RelExpr *node) {
	if (node->leftExpr()->precedence() <= node->precedence()) {
		if (isPrinting)
			out << "(";
		node->leftExpr()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->leftExpr()->accept(this);

	if (isPrinting)
		out << " ";
	rel_op(node->rOp);
	if (isPrinting)
		out << " ";

	if (node->rightExpr()->precedence() <= node->precedence()) {
		if (isPrinting)
			out << "(";
		node->rightExpr()->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->rightExpr()->accept(this);

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_cast(CastExpr *node) {
	if (isPrinting){
		out << "(";

		node->castTo->printType(out, NULL, true, 0);
	}
	if (isPrinting)
		out << ") ";

	if (isPrinting)
		out << "(";

	node->expr->accept(this);
	if (isPrinting)
		out << ")";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_sizeof(SizeofExpr *node) {
	if (isPrinting){
		out << "sizeof(";
	if (node->sizeofType)
		node->sizeofType->printType(out, NULL, true, 0);
	else
		node->expr->accept(this);
		out << ") ";
	}
	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_index(IndexExpr *node) {
	if (node->array->precedence() < node->precedence()) {
		if (isPrinting)
			out << "(";
		node->array->accept(this);
		if (isPrinting)
			out << ")";
	} else
		node->array->accept(this);

	if (isPrinting)
		out << "[";
	node->_subscript->accept(this);
	if (isPrinting)
		out << "]";

	/*#ifdef    SHOW_TYPES
	 if (node->type)
	 {
	 if (isPrinting)
	 out << openRemark;
	 node->type->printType(out, NULL, true, 0);
	 if (isPrinting)
	 out << closeOpenRemark;
	 }
	 #endif*/
}

void ParVarListTraversal::traverse_label(Label *node) {
	decr_indent();
	indent();
	incr_indent();

	switch (node->type) {
	case LT_None:
		assert(0);
		break;

	case LT_Default:
		if (isPrinting)
			out << "default";
		break;

	case LT_Case:
		assert(node->begin);
		if (isPrinting)
			out << "case ";
		node->begin->accept(this);
		break;

	case LT_CaseRange:
		assert(node->begin);
		assert(node->end);
		if (isPrinting)
			out << "case ";
		node->begin->accept(this);
		if (isPrinting)
			out << " ... ";
		node->end->accept(this);
		break;

	case LT_Goto:
		assert(node->name);
		traverse_symbol(node->name);
		break;
	}
	if (isPrinting)
		out << ":\n";
}

void ParVarListTraversal::traverse_decl(Decl *node) {

	if (isPrinting)
		node->print(out, show_base, indentLevel);

	ostringstream stam;
	node->print(stam, show_base, indentLevel);
}

void ParVarListTraversal::traverse_statement(Statement *node) {
	if (isPrinting)
		node->print(out, indentLevel);

	ostringstream stam;
	node->print(stam, indentLevel);

}

void ParVarListTraversal::traverse_file_line(FileLineStemnt *node) {
	/*#ifdef PRINT_LOCATION
	 if (isPrinting)
	 out << "$<FileLineStemnt:";
	 node->location.printLocation(out);
	 if (isPrinting)
	 out << ">$";
	 #endif*/
	if (node->linenumber > 0)
		if (isPrinting)
			out << "#line " << node->linenumber;
		else if (isPrinting)
			out << "#file";
	if (isPrinting)
		out << " \"" << node->filename << "\"" << std::endl;
}

void ParVarListTraversal::traverse_include(InclStemnt *node) {
	/*#ifdef PRINT_LOCATION
	 if (isPrinting)
	 out << "$<InclStemnt:";
	 node->location.printLocation(out);
	 if (isPrinting)
	 out << ">$";
	 #endif*/
	if (isPrinting)
		out << "#include " << (node->isStandard ? '<' : '"');
	if (isPrinting)
		out << node->filename << (node->isStandard ? '>' : '"') << std::endl;
}

void ParVarListTraversal::traverse_end_include(EndInclStemnt *node) {
	/*#ifdef PRINT_LOCATION
	 if (isPrinting)
	 out << "$<EndInclStemnt:";
	 node->location.printLocation(out);
	 if (isPrinting)
	 out << ">$";
	 #endif*/
}

void ParVarListTraversal::traverse_expression(ExpressionStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* ExpressionStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	assert(node->expression);
	node->expression->accept(this);
}

void ParVarListTraversal::traverse_if(IfStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* IfStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();

	if (isPrinting)
		out << "if (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")\n";
	block(node->thenBlk);

	if (node->elseBlk) {
		if (isPrinting)
			out << std::endl;
		indent();
		if (isPrinting)
			out << "else\n";
		block(node->elseBlk);
	}
}

void ParVarListTraversal::traverse_switch(SwitchStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* SwitchStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "switch (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
}

void ParVarListTraversal::traverse_for(ForStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* ForStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "regular for (";
	if (node->init)
		node->init->accept(this);
	if (isPrinting)
		out << "; ";

	if (node->cond)
		node->cond->accept(this);
	if (isPrinting)
		out << "; ";

	if (node->incr)
		node->incr->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
}

/* Gil Kulish added: */
void ParVarListTraversal::traverse_for(ParForStemnt *node) {

	activeInstance=this;
	if (debug) {
		if (isPrinting)
			out << "/* ParForStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "parfor" << "someName" << "();\n\n";

	if (isPrinting)
		out << "/*  this was originally a parfor,"
			"original code:\n  parfor (";
	if (node->init)
		node->init->accept(this);
	if (isPrinting)
		out << "; ";

	if (node->cond)
		node->cond->accept(this);
	if (isPrinting)
		out << "; ";

	if (node->incr)
		node->incr->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
	activeInstance=NULL;
}
/* end of Gil Kulish added: */

void ParVarListTraversal::traverse_while(WhileStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* WhileStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "while (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")\n";

	block(node->block);
}

void ParVarListTraversal::traverse_do_while(DoWhileStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* DoWhileStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "do ";

	if (!node->block->isBlock())
		if (isPrinting)
			out << std::endl;

	block(node->block);
	if (!node->block->isBlock())
		if (isPrinting)
			out << std::endl;

	indent();
	if (isPrinting)
		out << "while (";
	node->cond->accept(this);
	if (isPrinting)
		out << ")";
}

void ParVarListTraversal::traverse_goto(GotoStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* GotoStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "goto ";
	traverse_symbol(node->dest);
}

void ParVarListTraversal::traverse_return(ReturnStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* ReturnStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "return";

	if (node->result) {
		if (isPrinting)
			out << ' ';
		node->result->accept(this);
	}
}

void ParVarListTraversal::traverse_declaration(DeclStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* DeclStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (!node->decls.empty()) {
		DeclVector::const_iterator i = node->decls.begin();

		traverse_decl(*i);
		show_base = false;
		for (++i; i != node->decls.end(); ++i) {
			if (isPrinting)
				out << ", ";
			traverse_decl(*i);
		}
		show_base = true;
	}

	if (isPrinting)
		out << ";";
}

void ParVarListTraversal::traverse_typedef(TypedefStemnt *node) {
	if (debug) {
		if (isPrinting)
			out << "/* TypedefStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();

	if (isPrinting)
		out << "typedef ";

	if (!node->decls.empty()) {
		DeclVector::const_iterator i = node->decls.begin();
		traverse_decl(*i);
		for (++i; i != node->decls.end(); ++i) {
			if (isPrinting)
				out << ", ";
			traverse_decl(*i);
		}
	}

	if (isPrinting)
		out << ";";
}

void ParVarListTraversal::traverse_block(Block *node) {
	if (debug) {
		if (isPrinting)
			out << "/* BlockStemnt:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	for (LabelVector::const_iterator i = node->labels.begin(); i
			!= node->labels.end(); ++i)
		traverse_label(*i);

	indent();
	if (isPrinting)
		out << "{\n";

	bool isDecl = (node->head != 0) ? node->head->isDeclaration() : false;
	for (Statement *statement = node->head; statement; statement
			= statement->next) {
		if (isDecl && !statement->isDeclaration()) {
			isDecl = false;
			if (isPrinting)
				out << std::endl;
		}

		incr_indent();
		statement->accept(this);
		decr_indent();

		if (statement->needSemicolon())
			if (isPrinting)
				out << ";";
		if (isPrinting)
			out << std::endl;
	}
	indent();
	if (isPrinting)
		out << "}\n";
}

void ParVarListTraversal::traverse_function_definition(FunctionDef *node) {
	if (debug) {
		if (isPrinting)
			out << "/* FunctionDef:";
		node->location.printLocation(out);
		if (isPrinting)
			out << " */" << std::endl;
	}

	traverse_decl(node->decl);
	if (isPrinting)
		out << std::endl;

	traverse_block(node);
}

void ParVarListTraversal::traverse_unit(TransUnit *unit) {
	int inInclude = 0;
	for (Statement *statement = unit->head; statement; statement
			= statement->next) {
		if (inInclude > 0) {
			if (statement->isEndInclude())
				inInclude--;
			else if (statement->isInclude())
				inInclude++;
		} else {
			if (statement->isInclude())
				inInclude++;
			statement->accept(this);
			if (isPrinting)
				out << std::endl;
		}
	}
}

void ParVarListTraversal::meta_char(char c, bool in) {
	switch (c) {
	case '\'':
		if (in)
			if (isPrinting)
				out << "'";
			else if (isPrinting)
				out << "\\'";
		break;

	case '"':
		if (in)
			if (isPrinting)
				out << "\\\"";
			else if (isPrinting)
				out << "\"";
		break;

	case '\0':
		if (isPrinting)
			out << "\\0";
		break;
	case '\\':
		if (isPrinting)
			out << "\\\\";
		break;
	case '\n':
		if (isPrinting)
			out << "\\n";
		break;
	case '\t':
		if (isPrinting)
			out << "\\t";
		break;
	case '\r':
		if (isPrinting)
			out << "\\r";
		break;
	case '\f':
		if (isPrinting)
			out << "\\f";
		break;
	case '\b':
		if (isPrinting)
			out << "\\b";
		break;
	case '\v':
		if (isPrinting)
			out << "\\v";
		break;
	case '\a':
		if (isPrinting)
			out << "\\a";
		break;
	case ESC_VAL:
		if (isPrinting)
			out << "\\e";
		break;

	default:
		// Show low and high ascii as octal
		if ((c < ' ') || (c >= 127)) {
			char octbuf[8];
			sprintf(octbuf, "%03o", (unsigned char) c);
			if (isPrinting)
				out << "\\" << octbuf;
		} else if (isPrinting)
			out << c;
		break;
	}
}

void ParVarListTraversal::meta_string(const std::string &str) {
	for (unsigned i = 0; i < str.size(); ++i)
		meta_char(str[i], true);
}

void ParVarListTraversal::rel_op(RelOp op) {
	switch (op) {
	case RO_Equal:
		if (isPrinting)
			out << "==";
		break;

	case RO_NotEqual:
		if (isPrinting)
			out << "!=";
		break;

	case RO_Less:
		if (isPrinting)
			out << "<";
		break;

	case RO_LessEql:
		if (isPrinting)
			out << "<=";
		break;

	case RO_Grtr:
		if (isPrinting)
			out << ">";
		break;

	case RO_GrtrEql:
		if (isPrinting)
			out << ">=";
		break;
	}
}

void ParVarListTraversal::binary_op(BinaryOp op) {
	switch (op) {
	case BO_Plus:
		if (isPrinting)
			out << "+";
		break;

	case BO_Minus:
		if (isPrinting)
			out << "-";
		break;

	case BO_Mult:
		if (isPrinting)
			out << "*";
		break;

	case BO_Div:
		if (isPrinting)
			out << "/";
		break;

	case BO_Mod:
		if (isPrinting)
			out << "%";
		break;

	case BO_Shl:
		if (isPrinting)
			out << "<<";
		break;

	case BO_Shr:
		if (isPrinting)
			out << ">>";
		break;

	case BO_BitAnd:
		if (isPrinting)
			out << "&";
		break;

	case BO_BitXor:
		if (isPrinting)
			out << "^";
		break;

	case BO_BitOr:
		if (isPrinting)
			out << "|";
		break;

	case BO_And:
		if (isPrinting)
			out << "&&";
		break;

	case BO_Or:
		if (isPrinting)
			out << "||";
		break;

	case BO_Comma:
		if (isPrinting)
			out << ",";
		break;

	case BO_Member:
		if (isPrinting)
			out << ".";
		break;

	case BO_PtrMember:
		if (isPrinting)
			out << "->";
		break;

	default:
		//  case BO_Index        // x[y]
	case BO_Assign:
	case BO_Rel:
		break;
	}
}

void ParVarListTraversal::unary_op(UnaryOp op) {
	switch (op) {
	case UO_Plus:
		if (isPrinting)
			out << "+";
		break;

	case UO_Minus:
		if (isPrinting)
			out << "-";
		break;

	case UO_BitNot:
		if (isPrinting)
			out << "~";
		break;

	case UO_Not:
		if (isPrinting)
			out << "!";
		break;

	case UO_PreInc:
	case UO_PostInc:
		if (isPrinting)
			out << "++";
		break;

	case UO_PreDec:
	case UO_PostDec:
		if (isPrinting)
			out << "--";
		break;

	case UO_AddrOf:
		if (isPrinting)
			out << "&";
		break;

	case UO_Deref:
		if (isPrinting)
			out << "*";
		break;
	}
}

void ParVarListTraversal::assign_op(AssignOp op) {
	switch (op) {
	case AO_Equal:
		if (isPrinting)
			out << "=";
		break;

	case AO_PlusEql:
		if (isPrinting)
			out << "+=";
		break;

	case AO_MinusEql:
		if (isPrinting)
			out << "-=";
		break;

	case AO_MultEql:
		if (isPrinting)
			out << "*=";
		break;

	case AO_DivEql:
		if (isPrinting)
			out << "/=";
		break;

	case AO_ModEql:
		if (isPrinting)
			out << "%=";
		break;

	case AO_ShlEql:
		if (isPrinting)
			out << "<<=";
		break;

	case AO_ShrEql:
		if (isPrinting)
			out << ">>=";
		break;

	case AO_BitAndEql:
		if (isPrinting)
			out << "&=";
		break;

	case AO_BitXorEql:
		if (isPrinting)
			out << "^=";
		break;

	case AO_BitOrEql:
		if (isPrinting)
			out << "|=";
		break;
	}
}

void ParVarListTraversal::block(Statement *node) {
	if (!node) {
		incr_indent();
		null();
		decr_indent();
	} else if (node->isBlock())
		node->accept(this);
	else {
		incr_indent();
		node->accept(this);
		decr_indent();

		if (node->needSemicolon())
			if (isPrinting)
				out << ";";
	}
}
/**Gil Kulish:
 this method prints the declaration of a variable;
 it's copy-pasted from method void Decl::print, the only change is that the initializer, is not printed*/
void ParVarListTraversal::printDecl(Decl *decl, char pointerRef) {
	if (decl->form) {
		printType(decl->form,out, decl->name, true, 0,pointerRef);
	} else if (decl->name) {
		out << *decl->name;
	}

	if (decl->attrib) {
		decl->attrib->print(out);
	}

}

std::set<Decl*> * ParVarListTraversal::getVarList() {
	return &varList;
}

void
ParVarListTraversal::printType(Type *t, std::ostream& out, Symbol *name, bool showBase, int level, char pointerRef) const
{
    if (showBase)
	{
        t->printBase(out,level);

		if (name != NULL)
			out << " ";
	}

	string s("(");
	ostringstream ost;

    t->printBefore(ost,name,level);
	// fix the pointer type from "int (&*iden)" to "int (*&iden)":
	string fixPtr=ost.str();
	if (fixPtr.find("*")!=string::npos&&pointerRef=='&'){
		int location=fixPtr.find_last_of("*");
			fixPtr.replace(location,1,"*&");
	}else
		s+=pointerRef;
	if (pointerRef!=' ')
		out << s<<fixPtr<<")";
	else//set the output to be more readable; if not & or * needed
		out << fixPtr;
    t->printAfter(out);
}
