#include <ctool/PrintTraversalProxy.h>

/**the perpose of the class was to enable parsing of the precompiled par file,
 but printing the same non-pre-compiled parfile.
 in order to enable that, we need to print the non parsed file, until we encounter a
 function (with body), then we print it using the precompiled version of the par-file.

 TBD: one extension for this ability is to also handle typedefs in variable declerations,
 that are defined inside the header files.
 */
PrintTraversalProxy::PrintTraversalProxy(std::ostream &os, bool d,
		std::ostream &headerOut) :
	PrintTraversal(os, d, headerOut) {
}
void PrintTraversalProxy::setPreCompiledTUnit(TransUnit *preCompiledUnit) {
	this->preCompiledUnit = preCompiledUnit;
	nextPlaceToSearch = preCompiledUnit->head;
}
void PrintTraversalProxy::setPrintTraversal(PrintTraversal *pt) {
	this->pt = pt;
}

PrintTraversalProxy::~PrintTraversalProxy(void) {
}

void PrintTraversalProxy::traverse_function_definition(FunctionDef *node) {

	FunctionDef *preCompiledFunction = getFunctionFromPrecompiledUnit(node);
	pt->traverse_function_definition(preCompiledFunction);

}
/**
 * at some point in the compilation, there are two incarnations of the AST,
 * one that is not precompiled and one that is precompiled.
 * because the precompiled code contains a lot of unreadable code, we would prefer not to print this code
 * as the output code, but, since the precompiled code does not compile (since it is missing decelerations) we have to work in some
 * hybrid method.
 * so our algorithm to print the code is:
 *
 * work with the precompiled version (since it's the only true version the will compile),
 * !but! print to the output from the the version that wasn't precompiled.
 *
 * in order to do so, we present the following method.
 * the method returns un-precompiled version of a function.
 * the parameter passed to it is the precompiled function whom you want it's identical but not-precompiled version.
 */
FunctionDef* PrintTraversalProxy::getFunctionFromPrecompiledUnit(
		FunctionDef *func) {
	if (!func || !func->decl || !func->decl->name)
		return func;
	Statement* temp = nextPlaceToSearch;
	while (temp != NULL) {
		if (temp->isFuncDef()) {
			FunctionDef* function = ((FunctionDef*) temp);
			if (function->decl) {
				if (function->decl->name) {
					if (function->decl->name->name == func->decl->name->name) {
						nextPlaceToSearch = function;
						return function;
					}
				}
			}
		}
		temp = temp->next;
	}
	return func;
}

