#pragma once
#include <Runnable.h>
#include <list>
#include <Thread.h>
#include <pth.h>
#include <pth_p.h>
using namespace std;
#ifndef PUNIT_H

#define PUNIT_H

class ProcessingUnit : public Runnable{
public:
	/*the thread that is behind the ProcessingUnit*/
	Thread *t;
	ThreadLocalData *tld;
	static pthread_cond_t pUnitClassCond;
	static pthread_mutex_t pUnitClassMutex;
	pthread_mutex_t pUnitObjectMutex;
	pthread_cond_t pUnitObjectCond;
	/**
	 * used to allocate a unique sequential number to each and every
	 * ProcessingUnit object
	 */
	static int unique_processor_number_sequance;

	/**
	 * unique processor number of the ProcessingUnit
	 */
	int processorNumber;

	/**
	 * constructs a ProcessingUnit, creates a deamon thread in with all the
	 * tasks allocated to the ProcessingUnit via {@link #execute(Runnable)} will
	 * be executed.
	 *
	 * @see Thread#setDaemon(boolean)
	 */
public:
	ProcessingUnit();
	~ProcessingUnit();

	/**
	 * A queue that holds all the userActions that needs to be executed.
	 */
private:
	bool pth_check_no_threads_left();
	void debug_print_pth_threads_left();
	void pth_join_all();
	list<Runnable*> *userActionsToExecute;

	/**
	 * Inner class that contains the runnable method that defines the behaviour
	 * of ProcessingUnit's daemon thread. Generally speaking, the thread blocks
	 * until a <tt> Runnable </tt> object is inserted to it's
	 * {@link #userActionsToExecute} queue, then it awakes, executes all
	 * <tt> Runnable </tt> available in the queue and returens to wait again.
	 */
	void run();
public:
	void execute(Runnable *r);
	public:
	void join(Runnable* r);
public:
	void join();

};
#endif
