//#pragma once
#include <Runnable.h>
#include <vector>
#include <ProcessingUnit.h>
#include <Thread.h>
#include "globals.h"
using namespace std;
#ifndef PM_H
#define PM_H

class ProcessorsManager {
private:
	static vector<ProcessingUnit*> *processors;

	/**
	 * default number for active processes, even if the <tt>processors</tt>
	 * ArrayList is bigger, only the first <tt> activeProcessorsNumber </tt>
	 * processors will be used
	 *
	 * @see #activeProcessorsNumber
	 */
private:
	static const int DEFAULT_ACTIVE_PROCESSES_NUMBER;

	/**
	 * singleton
	 */
private:
	static ProcessorsManager *instance;

public:
	vector<ProcessingUnit*> *getProcessors();

	/**
	 * gets a specific ProcessingUnit <tt> n </tt> if the chosen ProcessingUnit
	 * number is out of bounds <code>(n>=activeProcessorsNumber)</code>, it
	 * returns the <code>n % activeProcessorsNumber</code> ProcessingUnit.
	 */
public:
	ProcessingUnit *getProcessor(int n);
	/**
	 * number of {@link ProcessingUnits} that should be used to execute actions.
	 */
private:
	int activeProcessorsNumber;

	/**
	 * number of {@link ProcessingUnits} been created and can be used (a pool)
	 * only {@link activeProcessorsNumber} will be used.
	 */
	int numberOfProcessorsInPool;
public:
	/* synchronized */void addProcessor() ;

	/* synchronized */
public:
	void addNProcessors(int n) ;
	/* synchronized */
public:
	void setNProcessors(int n) ;
	/* synchronized */
public:
	void removeProcessor();

private:
	ProcessorsManager();
public:
	virtual ~ProcessorsManager();
	/**
	 * as in {@link Thread#join()} blocks until all jobs of all processing units
	 * that have been posted up to this moment, have finished executing.
	 *
	 * @see Thread#join()
	 */
public:
	void execute(Runnable *r);

	void join();
	void wait_on_barrier(int barrier);
	/* singleton
	 */
public:
	static ProcessorsManager *getInstance();
	/* synchronized */
public:
	int getActiveProcessorsNumber();
};
#endif
